#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys, re, apt, threading, os
from PyQt4 import QtGui,QtCore,Qt
from PyQt4.QtCore import *
from leinstallpkgs.gui.interface import Ui_MainWindow
from leinstallpkgs.myprogress import Qt4AptProgress
from threading import Thread

class MyApt(apt.Cache,apt.Package):
    def __init__(self, progressopen=None):
        apt.Cache.__init__(self,progressopen)

    def getInstalledPackages(self, regex=None):
        """ Returns installed packages according to regex, or all installed
        packages if regex is None """

        if regex:
            regex = re.compile(regex)
            lf = lambda pkg: pkg.is_installed and regex.match(pkg.name)
        else:
            lf = lambda pkg: pkg.is_installed

        return [pkg.name for pkg in self if lf(pkg) ]

    def getAvailablePackages(self, regex=None):
        """ Returns available packages according to regex, or all available
        packages if regex is None """

        if regex:
            regex = re.compile(regex)
            lf = lambda pkg: regex.match(pkg.name)
            return [ pkg.name for pkg in self if lf(pkg) ]
        else:
            return self.keys()

class MainWindow(QtGui.QMainWindow, Ui_MainWindow):

    def __init__(self):
        QtGui.QMainWindow.__init__(self)

        # Set up the user interface from Designer
        self.setupUi(self)

        # Only root can install new packages
        if not self.imRoot():
            QtGui.QMessageBox.information(self,
                                u"Informação",
                                u"Você precisa ser root para instalar novos pacotes. O aplicativo irá iniciar com recursos limitados.")
            self.installPackagesButton.setEnabled(False)

        self.statusBarChangeMessage("Carregando lista de pacotes...", 0)

        # Progress Dialog
        self.progress = Qt4AptProgress(self)

        # Configuring widgets
        self.mainViewModel = QtGui.QStandardItemModel()
        self.viewAvailablePackages.setModel(self.mainViewModel)
        self.viewSelectedPackages.setModel(QtGui.QStandardItemModel())

        # Load packages list using a different thread
        Thread(None,self.loadAptCache).start()

    def imRoot(self):
        return os.geteuid() == 0 and True or False

    def loadAptCache(self):
        """ This method loads apt cache into self.__myapt var and fills the
        listview with the packages found. """

        self.__myapt = MyApt(self.progress.open())
        self.fillAvailablePackagesList()

        # Connect signals
        self.connectSignals()

        self.emit(QtCore.SIGNAL("messageChanged(PyQt_PyObject)"), "Pronto...")

    def statusBarChangeMessage(self, msg, timeout=2000):
        """ Called to change statusBar displaying message """

        self.statusBar.showMessage(msg, timeout)

    def connectSignals(self):
        """ Connect the signals that couldn't be connected on qtdesigner
        Tuple: (widget, signal, handler) """

        connectList = [(self,
                         "messageChanged(PyQt_PyObject)",
                         self.statusBarChangeMessage),
                       (self.searchButton,
                         "clicked()",
                         self.searchButtonClicked),
                       (self.viewAvailablePackages.selectionModel(),
                         "selectionChanged(QItemSelection,QItemSelection)",
                         self.viewAvailablePackagesSelected),
                       (self.addPackageButton,
                         "clicked()",
                         self.addPackageButtonClicked),
                       (self.rmPackageButton,
                         "clicked()",
                         self.rmPackageButtonClicked),
                       (self.installPackagesButton,
                         "clicked()",
                         self.installPackagesButtonClicked)
                      ]

        for widget,sig,handler in connectList:
            self.connect(widget,QtCore.SIGNAL(sig),handler)

    def viewAvailablePackagesSelected(self, selected, deselected):
        """ This method is called when the user selects a package from
        available packages list. The effect is print the package description
        into the description text widget. """

        items = self.getSelectedItemString()

        if (items and (len(items) == 1)):
            description = self.getPackageDescription(items[0])
        else:
            description = ""

        self.setPackageDescription(description)

    def setPackageDescription(self, description):
        """ Write description received into package description widget """

        self.packageDescription.setPlainText(description)

    def getPackageDescription(self, pkg):
        """ Returns the description of package pkg """

        return self.__myapt[pkg].description

    def fillAvailablePackagesList(self):
        """ Fill the availablePackagesList widget """

        for pkg in self.getMECPackages():
            self.addItemToModel(self.mainViewModel, pkg)

        self.sortModel(self.viewAvailablePackages.model())

    def sortModel(self, model):
        """ Sort a QListView alphabetically """

        model.sort(0)

    def findItemFromModel(self, model, text):
        """ Find an item named "text" from a QListView and return its index """

        matches = model.match(model.index(0, 0), Qt.DisplayRole, text);

        if (len(matches) > 0):
            return matches[0]
        else:
            return None

    def rmItemFromModel(self, model, text):
        """ Remove the item with the content "text" from a QListView """

        item = self.findItemFromModel(model, text);

        if item:
            model.removeRow(item.row())

    def addItemToModel(self, model, text, idx=None):
        """ Add an item named "text" to a model "model" at index "idx" """

        item = QtGui.QStandardItem(text)
        item.setEditable(False)

        if idx:
            model.setItem(idx, item)
        else:
            model.appendRow(item)

    def getDataFromIndexes(self, indexes):
        """ Return a list with text found in each index (probably from a
        selection on a listView) """

        return [ str(i.data().toString()) for i in indexes ]

    def getSelectedItem(self, listview=None):
        """ Returns selected item at ListView, or None if there is any """

        if listview == None:
            listview = self.viewAvailablePackages

        idxs = listview.selectedIndexes()
        if (len(idxs) > 0):
            return idxs
        else:
            return None

    def getSelectedItemString(self, model=None):
        """ Returns the string associated with the selected item at ListView,
        or None if there is any selected item """

        if model == None:
            model = self.viewAvailablePackages

        idxs = model.selectedIndexes()
        if (len(idxs) > 0):
            return self.getDataFromIndexes(idxs)
        else:
            return None

    def getMECPackages(self):
        """ Returns a list with all MEC packages available which if not
        installed on system  """

        regex = "mec-.*"
        available = self.__myapt.getAvailablePackages(regex)
        installed = self.__myapt.getInstalledPackages(regex)
        return [ i for i in available if i not in installed ]

    def getSearchParam(self):
        """ Return text typed on searchField formatted """

        return str(self.searchField.text()).strip().lower().decode()

    def searchPackages(self, param):
        """ Return packages that matchs param, using the name and
        description of it """

        model = self.mainViewModel
        result = []
        for i in range(model.rowCount()):
            pkgname = str(model.item(i).text()).lower()
            pkgdesc = self.getPackageDescription(pkgname).lower()
            if param in pkgname or param in pkgdesc:
                result.append(model.item(i))

        return result

    def searchButtonClicked(self):
        """ This method is called when the user clicks on "Search" button,
        and its purpose is find a package matching the given string """

        param = self.getSearchParam()

        if not param:
            newmodel = self.mainViewModel
        else:
            items = self.searchPackages(param)

            newmodel = QtGui.QStandardItemModel()
            for i in items:
                newmodel.appendRow(i.clone())

        self.viewAvailablePackages.setModel(newmodel)
        ism = QtGui.QItemSelectionModel(newmodel)
        self.viewAvailablePackages.setSelectionModel(ism)

        sig = "selectionChanged(QItemSelection,QItemSelection)"
        self.connect(self.viewAvailablePackages.selectionModel(),
                     QtCore.SIGNAL(sig),
                     self.viewAvailablePackagesSelected)

    def addPackageButtonClicked(self):
        """ Method called when user selects a new package to install """

        items = self.getSelectedItemString(self.viewAvailablePackages)
        for i in items:
            self.rmItemFromModel(self.mainViewModel, i)
            self.addItemToModel(self.viewSelectedPackages.model(), i)

        self.sortModel(self.viewSelectedPackages.model())

    def rmPackageButtonClicked(self):
        """ Method called when user removes a package from selection to
        install """

        items = self.getSelectedItemString(self.viewSelectedPackages)
        for i in items:
            self.rmItemFromModel(self.viewSelectedPackages.model(), i)
            self.addItemToModel(self.mainViewModel, i)

        self.sortModel(self.mainViewModel)

    def myapt(self):
        return self.__myapt

    def installPackagesButtonClicked(self):
        """ Method called when user clicks on install button """

        n = self.viewSelectedPackages.model().rowCount()
        for i in range(n):
            pkg = str(self.viewSelectedPackages.model().item(i).text())
            self.__myapt[pkg].mark_install()

        if self.__myapt.get_changes():
            f, i = self.progress.acquire(), self.progress.install()

            thread = MyAptCommitThread(self)
            thread.start()

            # Wait for dialog finish and check return result
            # If everything went ok, clear listview
            if self.progress.exec_() == self.progress.Accepted:
                for i in range(n):
                    self.viewSelectedPackages.model().removeRow(0)

class MyAptCommitThread(QtCore.QThread):
    """ Use a differente thread to commit apt changes """

    def __init__(self, parent):
        QtCore.QThread.__init__(self, parent)
        self.myapt = parent.myapt()
        self.progress = parent.progress

    def run(self):
        try:
            acquire = self.progress.acquire()
            install = self.progress.install()
            self.myapt.commit(acquire, install)
            # TODO: Only clear the list if the process went OK
            self.myapt.clear()
        except apt.cache.LockFailedException:
            text = "Aplicativo de instalação de pacotes já em uso, tente\
            novamente em alguns minutos."
            self.progress.error(text)
